@@ -5,6 +5,7 @@
  * $Author: kou $
  * $Date: 2008-09-26 14:13:58 $
  *
+ * Copyright 2003-2022 Sutou Kouhei <kou@cozmixng.org>
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
  *
@@ -17,10 +18,20 @@
 #include "rb_cairo_private.h"
 
 VALUE rb_cCairo_FontFace;
+VALUE rb_cCairo_FreeTypeFontFace = Qnil;
 VALUE rb_cCairo_ToyFontFace = Qnil;
 VALUE rb_cCairo_UserFontFace = Qnil;
 VALUE rb_cCairo_UserFontFace_TextToGlyphsData = Qnil;
 
+#ifdef CAIRO_HAS_FT_FONT
+#  include <cairo-ft.h>
+
+static cairo_user_data_key_t cr_freetype_face_key;
+static FT_Library cr_freetype_library;
+static int cr_freetype_n_faces = 0;
+static cairo_bool_t cr_freetype_finishing = CR_FALSE;
+#endif
+
 #if CAIRO_CHECK_VERSION(1, 7, 6)
 static cairo_user_data_key_t ruby_object_key;
 static ID cr_id_call;
@@ -28,6 +39,9 @@ static ID cr_id_new;
 
 static ID cr_id_init;
 static ID cr_id_render_glyph;
+#  if CAIRO_CHECK_VERSION(1, 17, 6)
+static ID cr_id_render_color_glyph;
+#endif
 static ID cr_id_text_to_glyphs;
 static ID cr_id_unicode_to_glyph;
 
@@ -84,6 +98,11 @@ rb_cairo_font_face_to_ruby_object (cairo_font_face_t *face)
 
       switch (cairo_font_face_get_type (face))
         {
+#ifdef CAIRO_HAS_FT_FONT
+        case CAIRO_FONT_TYPE_FT:
+          klass = rb_cCairo_FreeTypeFontFace;
+          break;
+#endif
 #if CAIRO_CHECK_VERSION(1, 7, 6)
         case CAIRO_FONT_TYPE_TOY:
           klass = rb_cCairo_ToyFontFace;
@@ -121,6 +140,147 @@ cr_font_face_quartz_supported_p (VALUE klass)
 #endif
 }
 
+static VALUE
+cr_font_face_freetype_supported_p (VALUE klass)
+{
+#ifdef CAIRO_HAS_FT_FONT
+  return Qtrue;
+#else
+  return Qfalse;
+#endif
+}
+
+#ifdef CAIRO_HAS_FT_FONT
+static VALUE
+cr_freetype_done_library (VALUE yielded_arg,
+                          VALUE callback_arg,
+                          int argc,
+                          const VALUE *argv,
+                          VALUE block_arg)
+{
+  cr_freetype_finishing = CR_TRUE;
+  if (cr_freetype_n_faces == 0)
+    FT_Done_FreeType (cr_freetype_library);
+  return Qnil;
+}
+
+static void
+cr_freetype_done_face (void *data)
+{
+  FT_Face face = data;
+
+  FT_Done_Face (face);
+  cr_freetype_n_faces--;
+  if (cr_freetype_n_faces == 0 && cr_freetype_finishing)
+    FT_Done_FreeType (cr_freetype_library);
+}
+
+static cairo_bool_t
+cr_freetype_error_detail (FT_Error error,
+                          const char **name,
+                          const char **message)
+{
+#undef __FTERRORS_H__
+#define FT_STRINGIFY(v) #v
+#define FT_ERRORDEF(e, v, s)    {FT_STRINGIFY(e), v, s},
+#define FT_ERROR_START_LIST     {
+#define FT_ERROR_END_LIST       {NULL, 0, NULL}};
+  size_t i = 0;
+  static const struct {
+    const char *name;
+    FT_Error code;
+    const char *message;
+  } errors[] =
+#include FT_ERRORS_H
+  for (i = 0; i < (sizeof (errors) / sizeof (errors[0])); i++)
+    {
+      if (errors[i].code == error)
+        {
+          *name = errors[i].name;
+          *message = errors[i].message;
+          return CR_TRUE;
+        }
+    }
+
+  return CR_FALSE;
+}
+
+static void
+cr_freetype_error_check (FT_Error error,
+                         const char *message,
+                         VALUE related_object)
+{
+  const char *name = NULL;
+  const char *system_message = NULL;
+  const char *error_class_name;
+  VALUE rb_eCairo_FreeTypeError;
+
+  if (error == FT_Err_Ok)
+    return;
+
+  cr_freetype_error_detail (error, &name, &system_message);
+#if CAIRO_CHECK_VERSION(1, 15, 4)
+  error_class_name = "FreeTypeError";
+#else
+  error_class_name = "Error";
+#endif
+  rb_eCairo_FreeTypeError =
+    rb_const_get (rb_mCairo, rb_intern (error_class_name));
+  if (NIL_P (related_object))
+    {
+      rb_raise (rb_eCairo_FreeTypeError,
+                "%s: %s[%d]: %s",
+                message,
+                name ? name : "unknown",
+                error,
+                system_message ? system_message : "unknown");
+    }
+  else
+    {
+      rb_raise (rb_eCairo_FreeTypeError,
+                "%s: %s[%d]: %s: %+" PRIsVALUE,
+                message,
+                name ? name : "unknown",
+                error,
+                system_message ? system_message : "unknown",
+                related_object);
+    }
+}
+
+static VALUE
+cr_freetype_font_face_initialize (VALUE self, VALUE path)
+{
+  FT_Face freetype_face;
+  FT_Error error;
+  cairo_font_face_t *face;
+  cairo_status_t status;
+
+  error = FT_New_Face (cr_freetype_library,
+                       StringValueCStr(path),
+                       0,
+                       &freetype_face);
+  cr_freetype_error_check (error, "failed to open FreeType font", path);
+  cr_freetype_n_faces++;
+
+  face = cairo_ft_font_face_create_for_ft_face (freetype_face, 0);
+  cr_font_face_check_status (face);
+  status =
+    cairo_font_face_set_user_data (face,
+                                   &cr_freetype_face_key,
+                                   freetype_face,
+                                   (cairo_destroy_func_t) cr_freetype_done_face);
+  if (status != CAIRO_STATUS_SUCCESS) {
+    cairo_font_face_destroy (face);
+    FT_Done_Face (freetype_face);
+    rb_cairo_check_status (status);
+  }
+
+  DATA_PTR (self) = face;
+
+  return Qnil;
+}
+#endif
+
 #if CAIRO_CHECK_VERSION(1, 7, 6)
 static VALUE
 cr_toy_font_face_initialize (int argc, VALUE *argv, VALUE self)
@@ -306,10 +466,11 @@ cr_user_font_face_render_glyph_func_after (VALUE user_data)
 }
 
 static cairo_status_t
-cr_user_font_face_render_glyph_func (cairo_scaled_font_t *scaled_font,
-                                     unsigned long glyph,
-                                     cairo_t *cr,
-                                     cairo_text_extents_t *extents)
+cr_user_font_face_render_glyph_func_internal (cairo_scaled_font_t *scaled_font,
+                                              unsigned long glyph,
+                                              cairo_t *cr,
+                                              cairo_text_extents_t *extents,
+                                              ID id_registered_method_name)
 {
   cairo_status_t status = CAIRO_STATUS_SUCCESS;
   cairo_font_face_t *face;
@@ -320,10 +481,11 @@ cr_user_font_face_render_glyph_func (cairo_scaled_font_t *scaled_font,
   face = cairo_scaled_font_get_font_face (scaled_font);
   self = (VALUE)cairo_font_face_get_user_data (face, &ruby_object_key);
   receiver = rb_ivar_get (self, cr_id_render_glyph);
-  if (NIL_P (receiver) && rb_obj_respond_to (self, cr_id_render_glyph, Qtrue))
+  if (NIL_P (receiver) &&
+      rb_obj_respond_to (self, id_registered_method_name, Qtrue))
     {
       receiver = self;
-      id_method_name = cr_id_render_glyph;
+      id_method_name = id_registered_method_name;
     }
 
   if (!NIL_P (receiver))
@@ -350,6 +512,34 @@ cr_user_font_face_render_glyph_func (cairo_scaled_font_t *scaled_font,
   return status;
 }
 
+static cairo_status_t
+cr_user_font_face_render_glyph_func (cairo_scaled_font_t *scaled_font,
+                                     unsigned long glyph,
+                                     cairo_t *cr,
+                                     cairo_text_extents_t *extents)
+{
+  return cr_user_font_face_render_glyph_func_internal (scaled_font,
+                                                       glyph,
+                                                       cr,
+                                                       extents,
+                                                       cr_id_render_glyph);
+}
+
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+static cairo_status_t
+cr_user_font_face_render_color_glyph_func (cairo_scaled_font_t *scaled_font,
+                                           unsigned long glyph,
+                                           cairo_t *cr,
+                                           cairo_text_extents_t *extents)
+{
+  return cr_user_font_face_render_glyph_func_internal (scaled_font,
+                                                       glyph,
+                                                       cr,
+                                                       extents,
+                                                       cr_id_render_color_glyph);
+}
+#endif
+
 typedef struct _cr_text_to_glyphs_after_hook_data {
   VALUE text_to_glyphs_data;
   cairo_glyph_t **glyphs;
@@ -546,6 +736,10 @@ cr_user_font_face_initialize (VALUE self)
     (face, cr_user_font_face_init_func);
   cairo_user_font_face_set_render_glyph_func
     (face, cr_user_font_face_render_glyph_func);
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+  cairo_user_font_face_set_render_color_glyph_func
+    (face, cr_user_font_face_render_color_glyph_func);
+#endif
   cairo_user_font_face_set_text_to_glyphs_func
     (face, cr_user_font_face_text_to_glyphs_func);
   cairo_user_font_face_set_unicode_to_glyph_func
@@ -575,6 +769,15 @@ cr_user_font_face_on_render_glyph (VALUE self)
   return self;
 }
 
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+static VALUE
+cr_user_font_face_on_render_color_glyph (VALUE self)
+{
+  rb_ivar_set (self, cr_id_render_color_glyph, rb_block_proc ());
+  return self;
+}
+#endif
+
 static VALUE
 cr_user_font_face_on_text_to_glyphs (VALUE self)
 {
@@ -647,6 +850,9 @@ Init_cairo_font (void)
 
   cr_id_init = rb_intern ("init");
   cr_id_render_glyph = rb_intern ("render_glyph");
+#  if CAIRO_CHECK_VERSION(1, 17, 6)
+  cr_id_render_color_glyph = rb_intern ("render_color_glyph");
+#endif
   cr_id_text_to_glyphs = rb_intern ("text_to_glyphs");
   cr_id_unicode_to_glyph = rb_intern ("unicode_to_glyph");
 
@@ -664,6 +870,26 @@ Init_cairo_font (void)
 
   rb_define_singleton_method (rb_cCairo_FontFace, "quartz_supported?",
                               cr_font_face_quartz_supported_p, 0);
+  rb_define_singleton_method (rb_cCairo_FontFace, "freetype_supported?",
+                              cr_font_face_freetype_supported_p, 0);
+
+#ifdef CAIRO_HAS_FT_FONT
+  rb_cCairo_FreeTypeFontFace =
+    rb_define_class_under (rb_mCairo, "FreeTypeFontFace", rb_cCairo_FontFace);
+
+  {
+    FT_Error error;
+
+    error = FT_Init_FreeType (&cr_freetype_library);
+    cr_freetype_error_check (error, "failed to initialize FreeType", Qnil);
+
+    rb_define_finalizer (rb_cCairo_FreeTypeFontFace,
+                         rb_proc_new (cr_freetype_done_library, Qnil));
+  }
+
+  rb_define_method (rb_cCairo_FreeTypeFontFace, "initialize",
+                    cr_freetype_font_face_initialize, 1);
+#endif
 
 #if CAIRO_CHECK_VERSION(1, 7, 6)
   rb_cCairo_ToyFontFace =
@@ -690,6 +916,10 @@ Init_cairo_font (void)
                     cr_user_font_face_on_init, 0);
   rb_define_method (rb_cCairo_UserFontFace, "on_render_glyph",
                     cr_user_font_face_on_render_glyph, 0);
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+  rb_define_method (rb_cCairo_UserFontFace, "on_render_color_glyph",
+                    cr_user_font_face_on_render_color_glyph, 0);
+#endif
   rb_define_method (rb_cCairo_UserFontFace, "on_text_to_glyphs",
                     cr_user_font_face_on_text_to_glyphs, 0);
   rb_define_method (rb_cCairo_UserFontFace, "on_unicode_to_glyph",
