@@ -2,7 +2,7 @@
 /*
  * Ruby Cairo Binding
  *
- * Copyright 2005-2018 Kouhei Sutou <kou@cozmixng.org>
+ * Copyright 2005-2022 Sutou Kouhei <kou@cozmixng.org>
  * Copyright 2014 Patrick Hanevold <patrick.hanevold@gmail.com>
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
@@ -86,7 +86,6 @@ VALUE rb_cCairo_GLTextureSurface = Qnil;
 VALUE rb_cCairo_DRMSurface = Qnil;
 VALUE rb_cCairo_TeeSurface = Qnil;
 VALUE rb_cCairo_XMLSurface = Qnil;
-VALUE rb_cCairo_SkiaSurface = Qnil;
 VALUE rb_cCairo_SubSurface = Qnil;
 VALUE rb_cCairo_CoglSurface = Qnil;
 
@@ -195,9 +194,6 @@ cr_surface_get_klass (cairo_surface_t *surface)
     case CAIRO_SURFACE_TYPE_XML:
       klass = rb_cCairo_XMLSurface;
       break;
-    case CAIRO_SURFACE_TYPE_SKIA:
-      klass = rb_cCairo_SkiaSurface;
-      break;
     case CAIRO_SURFACE_TYPE_SUBSURFACE:
       klass = rb_cCairo_SubSurface;
       break;
@@ -236,6 +232,28 @@ rb_cairo_surface_adjust_memory_usage (cairo_surface_t *surface,
 #endif
 }
 
+static cairo_surface_t *
+rb_cairo_surface_from_ruby_object_without_null_check (VALUE obj)
+{
+  cairo_surface_t *surface;
+  if (!rb_cairo__is_kind_of (obj, rb_cCairo_Surface))
+    {
+      rb_raise (rb_eTypeError, "not a cairo surface");
+    }
+  Data_Get_Struct (obj, cairo_surface_t, surface);
+  return surface;
+}
+
+cairo_surface_t *
+rb_cairo_surface_from_ruby_object (VALUE obj)
+{
+  cairo_surface_t *surface =
+    rb_cairo_surface_from_ruby_object_without_null_check (obj);
+  if (!surface)
+    rb_cairo_check_status (CAIRO_STATUS_NULL_POINTER);
+  return surface;
+}
+
 static void
 cr_surface_destroy_raw (cairo_surface_t *surface)
 {
@@ -258,7 +276,7 @@ cr_surface_destroy (VALUE self)
 static VALUE
 cr_surface_destroy_with_destroy_check (VALUE self)
 {
-  if (_SELF)
+  if (rb_cairo_surface_from_ruby_object_without_null_check (self))
     cr_surface_destroy (self);
   return Qnil;
 }
@@ -411,21 +429,6 @@ cr_surface_xml_supported_p (VALUE klass)
 #endif
 }
 
-/* constructor/de-constructor */
-cairo_surface_t *
-rb_cairo_surface_from_ruby_object (VALUE obj)
-{
-  cairo_surface_t *surface;
-  if (!rb_cairo__is_kind_of (obj, rb_cCairo_Surface))
-    {
-      rb_raise (rb_eTypeError, "not a cairo surface");
-    }
-  Data_Get_Struct (obj, cairo_surface_t, surface);
-  if (!surface)
-    rb_cairo_check_status (CAIRO_STATUS_NULL_POINTER);
-  return surface;
-}
-
 static rb_cairo__object_holder_t *
 cr_object_holder_new (VALUE object)
 {
@@ -519,6 +522,7 @@ cr_surface_finish (VALUE self)
   cairo_surface_finish (surface);
   cairo_surface_set_user_data (surface, &cr_finished_key, (void *)CR_TRUE, NULL);
   cairo_surface_set_user_data (surface, &cr_object_holder_key, NULL, NULL);
+  DATA_PTR (self) = NULL;
 
   if (closure && !NIL_P (closure->error))
     rb_exc_raise (closure->error);
@@ -535,7 +539,7 @@ rb_cairo__surface_yield_and_finish (VALUE self)
 
   rb_result = rb_yield (self);
 
-  surface = _SELF;
+  surface = rb_cairo_surface_from_ruby_object_without_null_check (self);
   if (!surface)
     return rb_result;
   if (cairo_surface_status (surface) != CAIRO_STATUS_SUCCESS)
@@ -1332,6 +1336,26 @@ cr_pdf_surface_set_thumbnail_size (VALUE self,
   return Qnil;
 }
 #  endif
+
+#  if CAIRO_CHECK_VERSION(1, 17, 6)
+static VALUE
+cr_pdf_surface_set_custom_metadata (VALUE self,
+                                    VALUE rb_name,
+                                    VALUE rb_value)
+{
+  cairo_surface_t *surface;
+  const char *name;
+  const char *value;
+
+  surface = _SELF;
+  name = RVAL2CSTR (rb_name);
+  value = RVAL2CSTR (rb_value);
+  cairo_pdf_surface_set_custom_metadata (surface, name, value);
+  rb_cairo_surface_check_status (surface);
+
+  return Qnil;
+}
+#  endif
 #endif
 
 #ifdef CAIRO_HAS_PS_SURFACE
@@ -2111,9 +2135,9 @@ Init_cairo_surface (void)
 #endif
 #if CAIRO_CHECK_VERSION(1, 5, 2)
   rb_define_method (rb_cCairo_Surface, "copy_page",
-                    cr_surface_copy_page, 2);
+                    cr_surface_copy_page, 0);
   rb_define_method (rb_cCairo_Surface, "show_page",
-                    cr_surface_show_page, 2);
+                    cr_surface_show_page, 0);
 #endif
 
 #ifdef CAIRO_HAS_PNG_FUNCTIONS
@@ -2181,6 +2205,11 @@ Init_cairo_surface (void)
   }
 #  endif
 
+#  if CAIRO_CHECK_VERSION(1, 17, 6)
+  rb_define_method (rb_cCairo_PDFSurface, "set_custom_metadata",
+                    cr_pdf_surface_set_custom_metadata, 2);
+#  endif
+
   RB_CAIRO_DEF_SETTERS (rb_cCairo_PDFSurface);
 #endif
 
@@ -2358,10 +2387,6 @@ Init_cairo_surface (void)
   RB_CAIRO_DEF_SETTERS (rb_cCairo_XMLSurface);
 #endif
 
-  /* Skia surface */
-  rb_cCairo_SkiaSurface =
-    rb_define_class_under (rb_mCairo, "SkiaSurface", rb_cCairo_Surface);
-
   /* sub surface */
   rb_cCairo_SubSurface =
     rb_define_class_under (rb_mCairo, "SubSurface", rb_cCairo_Surface);
